#!/usr/bin/env perl
#CVS: $Id: backspace_sim.pl,v 1.5 2010/05/05 21:03:21 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [backspace_sim.pl]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [This is the main BackSpace script                ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
use 5.008;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;
use ExtractVars;
use GenerateSymTable;
use GenerateDimacs;
use DCverPrettyCode;
use Sim2Sat;
use GenerateInstance;

use Benchmark;

#
# Process Input Params
#
my $help      = 0;     # option variable with default value (false)
my $ncycles   = 1;     # option variable defining # of cycles to run
my $resudir   = "RESULTS";      # option variable defining results directory
my $inFile    = undef;    # Techmapped verilog file name 
my $nBckSpc   = 5000; # number of cycles to backspace
my $local = "";    # Techmapped verilog file name 

GetOptions ('help|?'    => \$help,
            'ncycles=i' => \$ncycles,
            'resdir=s'  => \$resudir,
            'inFile=s'  => \$inFile,
	    'nBkSpc=i'  => \$nBckSpc,
	    'local=s' => \$local
	    );
pod2usage(-exitval => 0, -verbose => 2) if $help;

########################################################################
# Variable declarations and definitions
########################################################################
#
# File management objects
#
my $obj         = new MyFileHandlers;
my $files_exist = new MyFileHandlers;

#
# Pre-processing objects
#
my $prettyCode = new DCverPrettyCode;
my $symbTable  = new GenerateSymTable;
my $dimacs     = new GenerateDimacs;

#
# Global structures
#
my @cleaned_file = ();
my $symTable_ref = {};
my @clauses      = ();
 
# Used for user-interaction
my $answ ='';

#
# Misc variables
#
my $i          = 0; 
my $sat_status = 0;
my $status     = 1;
my $file       = '';
my @date;
my $ssh_cmd    = '';
# Used for benchmarking
my $t0; 
my $t1;
my $td;


########################################################################
# Initial environment setup
########################################################################
#
# Creates absolute paths and check that it exists
#
my $currdir = $obj->MFHcurdir;
print "BACSPACE INFO: \$CWD =  $currdir\n";
my $resdir = $obj->MFHconcatdir2($currdir,$resudir);
my $workdir = $obj->MFHconcatdir2($currdir,".work");
my $srcdir  = "./";

# VCS executable
# File that contains the crash and pre-crash states
# File that contains a loadable breakpoint 
system("ln -s ../sim/simv") unless -e "simv"; 
my $simv           = $obj->MFHconcatfile($currdir,"simv");
my $pair_of_states = $obj->MFHconcatfile($currdir,"states.log");
my $breakpoint     = $obj->MFHconcatfile($currdir,"breakpoint");

# Identify where states.log lives
# Identify where the sw running on the DUV lives
my $states = $obj->MFHconcatdir2($currdir,"STATES");
system("cd $states; ln -s ../states.log") unless -e "$states/states.log"; 
system("ln -s ../sim/sw-images") unless -e "sw-images";

#
# Used in the past; Now, default to 1 if no seed is passed
my $seed = 1;
srand(time());

#########################################################################
## Initial Setup 
#########################################################################
if( -e ".work/cleaned.v"){
    print "BACKSPACE: Warning:: .work dir already exists; Do you want to overwrite all files \'[y|n] ENTER\'?";
    read STDIN, $answ, 2;
    if($answ =~ /^y/i){
	if(!defined($inFile)) {
	    die "Need to pass paramemter --inFile pointing to techmapped verilog file...Exiting\n";
	}
	clean_work(); #system("rm .work/state_filter .work/constraint_indexes .work/reset_ID");
	$obj->MFHreadfile("$inFile");
	my @file_bare_bones = $obj->MFHfile;
	@cleaned_file = $prettyCode->cleanUpCode(\@file_bare_bones);
	$symTable_ref = $symbTable->getVariables(\@cleaned_file);
	@clauses =  $dimacs->generateDimacs(\@cleaned_file, $symTable_ref);
    }else{
	#### READ in files
	$files_exist->MFHreadfile(".work/cleaned.v");
	@cleaned_file = $files_exist->MFHfile;

	$files_exist->MFHreadfile(".work/symTable");
	my @to_hash = $files_exist->MFHfile;
	$symbTable->readHash(\@to_hash, $symTable_ref);

	$files_exist->MFHreadfile(".work/clauses");
	@clauses = $files_exist->MFHfile;
    }
}else{
    if(!defined($inFile)) {
	die "Need to pass paramemter --inFile pointing to techmapped verilog file...Exiting\n";
    }
    clean_work(); #system("rm .work/state_filter .work/constraint_indexes .work/reset_ID");
    $obj->MFHreadfile("$inFile");
    my @file_bare_bones = $obj->MFHfile;
    @cleaned_file = $prettyCode->cleanUpCode(\@file_bare_bones);
    $symTable_ref = $symbTable->getVariables(\@cleaned_file);
    @clauses =  $dimacs->generateDimacs(\@cleaned_file, $symTable_ref);
}

if($local =~ /cs/i){
    # Used to be necessary to ssh to a different machine (platform to run vcs)
    #$ssh_cmd = "ssh columbia \'setenv VCS_LIC_EXPIRE_WARNING 5;";
    $ssh_cmd = "set VCS_LIC_EXPIRE_WARNING 5;";
}else{
    die "Define which machine VCS is located... Exiting"; 
    $ssh_cmd = "";
}


#########################################################################
## MAIN CODE 
#########################################################################
#
# Backspace Loop 
#
while($i<$nBckSpc){
    if(-x $simv){
	@date = localtime(time);
	print "----------------->>>\n";
	print "Run $i : Seed  $seed   Time: $date[2]:$date[1]:$date[0]\n";
	print ">>>>>---------------\n";
	print ".Running logic simulation\n";
	if($i == 0){
	    system("$ssh_cmd cd $currdir;simv +vcs+lic+wait +seed=$seed +ncycles=$ncycles > $resdir/trace.i$i.n$ncycles.s$seed");
	    $file = "$resdir/trace.i$i.n$ncycles.s$seed";
	    $status = parseOutput($file);
	}else{
	    system ("$ssh_cmd cd $currdir;simv +vcs+lic+wait +seed=$seed +read_bp=1 > $resdir/trace.i$i.n$ncycles.s$seed");
	    $file = "$resdir/trace.i$i.n$ncycles.s$seed";
	    $status = parseOutput($file);
	}
	if($status == 1){
	    if(! -z $pair_of_states){
		minisatInfra(\@cleaned_file, \@clauses, $symTable_ref);
		$sat_status = callMisat();
		if($sat_status == 0) {
		    print "TERMINATION CONDITION: ERROR: Found no assignments in run $i\n";
		    exit(0);
		}
		if(($sat_status >= 2) && ($sat_status <=300)) {
		    findCorrectAssignment($sat_status);
		}
		if($sat_status > 300){
		    print "TERMINATION CONDITION: $sat_status exceeds limit of assignments of 300 in run $i\n";
		    exit(0);
		}
	    }
	}elsif($status == 2){	 
	    print "ERROR  : Exiting backspace framework...\n";
	    exit(0);
	}elsif($status == 0){	 
	    print "SUCCESS: Exiting backspace framework...\n";
	    exit(0);
	}
    }else{
	die "BACKSPACE: Error: simulator does not exist...Exiting\n";
    }
    $i = $i + 1;

}
print "Bounded Sucess: backspaced $nBckSpc cycles... Exiting ...\n";
exit(0);

###########################################################################
# SUBROUTINES
###########################################################################
#
#
#
sub minisatInfra{
    my $verilog_ptr = shift;
    my $clauses_ptr = shift;
    my $symbTbl_ptr = shift;

    my $sim2Sat    = new Sim2Sat;
    my $problem    = new GenerateInstance;

    my $design_path = '.';
    my $vcsdir    = './STATES';     
    my $oudir     = 'two_states';      
    my $stateslog = 'states.log';        
    my $dontcares = "$design_path/dontcares";
    
    print ". . Parsing crash state\n";
    $sim2Sat->generateFiles($vcsdir,$oudir,$stateslog,$dontcares, $verilog_ptr);
    print ". . . Generating SAT Problem Instance\n";
    $problem->instantiate($clauses_ptr, $symbTbl_ptr, $verilog_ptr,"./two_states/STATES/setOfConstraints.h", "./two_states/STATES/setOfStates.h","./two_states/STATES/careSet.h");
}
#
#
#
sub callMisat{
    my $FH;
    my $status = 0;
    my $line;
    my $SAT = $currdir . "/SAT";
    print ". . . . . Computing pre-image\n";
    system("mkdir $SAT/LOGS") unless -e "$SAT/LOGS";
    system("cd $SAT; ln -s ../cnfLatches.h") unless -e "$SAT/cnfLatches.h";
    system("cd $SAT; ln -s ../constraints0.h constraints.h") unless -e "$SAT/constraints.h";
    system("cd $SAT; gmake clean; gmake;") unless -e "$SAT/minisat";
    system("cd $SAT; minisat -alpha=0.5 -verbosity=0 -seed=0 ../instance0.cnf >& LOGS/status$i.log");
    open($FH, "<SAT/LOGS/status$i.log") || die "Cannot read SAT/LOGS/status$i.log";
    while(defined($line=readline($FH))){
	if($line =~ /Found\s1\ssatisfying assignments/){
	    $status = 1;
	    last;
	}
	if($line =~ /Found\s([0-9]+).*Multiple assignments/){
	    $status = $1;
	    last;
	}
    }
    return $status;
}
#
#
#
sub findCorrectAssignment{
    my $upperBound = shift;
    print "===> Found multiple states during the pre-image computation. Total is $upperBound.\n     Determining the correct one...\n";
    my $i=0;
    my $FH;
    my $line;
    my @state;
    my $instance;
    my $file;
    my $status;
    my $timeout = $ncycles + 1;
    system("cp breakpoint break2");
    open($FH,"<breakpoint") || die "Cannot read breakpoint file\n";
    while(defined($line=readline($FH))){
	chomp $line;
	push @state, $line;
    }
    close($FH);
    # assertion
    if($upperBound != ($#state + 1)) {
	print "ERROR: Read $#state + 1 instead of $upperBound\n";
	exit(0);
    }

    foreach $instance (@state) {
	print "===> Iteration $i\n";
	open($FH,">breakpoint") || die "Cannot write to breakpoint file\n";
	print $FH $instance;
	close($FH);
	## Replace ssh columbia ; FIX ME
	system ("$ssh_cmd cd $currdir; simv +vcs+lic+wait +seed=$seed +read_bp=1 +timeout=$timeout > $resdir/search.i$i.n$ncycles.s$seed\'");
	$file = "$resdir/search.i$i.n$ncycles.s$seed";
	$status = parseOutput($file);
	print "status of search is $status\n";
	if ($status != 3){
	    print "===> Found correct assignment\n";
	    last;
	}
	$i = $i + 1;
    }
    if($status == 3){	 
	print "ERROR  : Timed out over all states in pre-image...\n";
	exit(0);
    }elsif($status == 2){	 
	print "ERROR  : Exiting backspace framework...\n";
	exit(0);
    }elsif($status == 0){	 
	print "SUCCESS: Exiting backspace framework...\n";
	exit(0);
    } 

}
#
#
#
sub parseOutput{
    my $file = shift;
    my $FH;
    my $line;
    my $status = 2;
    open($FH,"<$file") || die "Cannot read file $file\n";
    
    while(defined($line = readline($FH))){
	if($line =~ /Timed out/) {
	    $status = 3;
	    last;
	}elsif($line =~ /error/i) {
	    $status = 2;
	    last;
	}elsif($line =~ /finishing/){
	    $status = 1;
	    last;
	}elsif($line =~ /exiting/){
	    $status = 0;
	   last;
	}
    }
    close($FH);
    return $status;
}

sub clean_work{

    if(-e ".work/constraint_indexes"){system("rm .work/constraint_indexes");}
    if(-e ".work/state_filter")      {system("rm .work/state_filter");}
    if(-e ".work/reset_ID")          {system("rm .work/reset_ID");}
    if(-e ".work/symTable")          {system("rm .work/symTable");}
    if(-e ".work/clauses")           {system("rm .work/clauses");}
    if(-e ".work/cleaned.v")         {system("rm .work/cleaned.v");}
}
