/******************************************************************************************[Main.C]
#CVS: $Id: Main.C,v 1.2 2009/03/30 11:37:34 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [Main.C]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [API to MiniSat; built on top of the distributed  ]
#                             [Minisat 2.0. Each design will have it's own copy ]
#                             [this file.                                       ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
MiniSat -- Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/
#include <iostream>
#include <string>
#include <fstream>
#include <vector>
using namespace std;

#include <ctime>
#include <cstring>
#include <stdint.h>
#include <errno.h>

#include <signal.h>
#include <zlib.h>

#include "Solver.h"

#include <time.h>
#include <assert.h>

// BACKSPACE defines
#define BACKSPACEABILITY 300
#define BACKSPACE_PRINT_STEP 100


/*************************************************************************************/
#ifdef _MSC_VER
#include <ctime>

static inline double cpuTime(void) {
    return (double)clock() / CLOCKS_PER_SEC; }
#else

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

static inline double cpuTime(void) {
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; }
#endif


#if defined(__linux__)
static inline int memReadStat(int field)
{
    char    name[256];
    pid_t pid = getpid();
    sprintf(name, "/proc/%d/statm", pid);
    FILE*   in = fopen(name, "rb");
    if (in == NULL) return 0;
    int     value;
    for (; field >= 0; field--)
        fscanf(in, "%d", &value);
    fclose(in);
    return value;
}
static inline uint64_t memUsed() { return (uint64_t)memReadStat(0) * (uint64_t)getpagesize(); }


#elif defined(__FreeBSD__)
static inline uint64_t memUsed(void) {
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return ru.ru_maxrss*1024; }


#else
static inline uint64_t memUsed() { return 0; }
#endif

#if defined(__linux__)
#include <fpu_control.h>
#endif

//=================================================================================================
// DIMACS Parser:

#define CHUNK_LIMIT 1048576

class StreamBuffer {
    gzFile  in;
    char    buf[CHUNK_LIMIT];
    int     pos;
    int     size;

    void assureLookahead() {
        if (pos >= size) {
            pos  = 0;
            size = gzread(in, buf, sizeof(buf)); } }

public:
    StreamBuffer(gzFile i) : in(i), pos(0), size(0) {
        assureLookahead(); }

    int  operator *  () { return (pos >= size) ? EOF : buf[pos]; }
    void operator ++ () { pos++; assureLookahead(); }
};

//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template<class B>
static void skipWhitespace(B& in) {
    while ((*in >= 9 && *in <= 13) || *in == 32)
        ++in; }

template<class B>
static void skipLine(B& in) {
    for (;;){
        if (*in == EOF || *in == '\0') return;
        if (*in == '\n') { ++in; return; }
        ++in; } }

template<class B>
static int parseInt(B& in) {
    int     val = 0;
    bool    neg = false;
    skipWhitespace(in);
    if      (*in == '-') neg = true, ++in;
    else if (*in == '+') ++in;
    if (*in < '0' || *in > '9') reportf("parseInt: PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
    while (*in >= '0' && *in <= '9')
        val = val*10 + (*in - '0'),
        ++in;
    return neg ? -val : val; }

template<class B>
static void readClause(B& in, Solver& S, vec<Lit>& lits) {
    int     parsed_lit, var;
    lits.clear();
    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        while (var >= S.nVars()) S.newVar();
        lits.push( (parsed_lit > 0) ? Lit(var) : ~Lit(var) );
    }
}

template<class B>
static bool match(B& in, char* str) {
    for (; *str != 0; ++str, ++in)
        if (*str != *in)
            return false;
    return true;
}


template<class B>
static void parse_DIMACS_main(B& in, Solver& S) {
    vec<Lit> lits;
    for (;;){
        skipWhitespace(in);
        if (*in == EOF)
            break;
        else if (*in == 'p'){
            if (match(in, "p cnf")){
                int vars    = parseInt(in);
                int clauses = parseInt(in);
                reportf("|  Number of variables:  %-12d                                         |\n", vars);
                reportf("|  Number of clauses:    %-12d                                         |\n", clauses);
            }else{
                reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
            }
        } else if (*in == 'c' || *in == 'p')
            skipLine(in);
        else
            readClause(in, S, lits),
            S.addClause(lits);
    }
}

// Inserts problem into solver.
//
static void parse_DIMACS(gzFile input_stream, Solver& S) {
    StreamBuffer in(input_stream);
    parse_DIMACS_main(in, S); }


//=================================================================================================


void printStats(Solver& solver)
{
    double   cpu_time = cpuTime();
    uint64_t mem_used = memUsed();
    reportf("restarts              : %lld\n", solver.starts);
    reportf("conflicts             : %-12lld   (%.0f /sec)\n", solver.conflicts   , solver.conflicts   /cpu_time);
    reportf("decisions             : %-12lld   (%4.2f %% random) (%.0f /sec)\n", solver.decisions, (float)solver.rnd_decisions*100 / (float)solver.decisions, solver.decisions   /cpu_time);
    reportf("propagations          : %-12lld   (%.0f /sec)\n", solver.propagations, solver.propagations/cpu_time);
    reportf("conflict literals     : %-12lld   (%4.2f %% deleted)\n", solver.tot_literals, (solver.max_literals - solver.tot_literals)*100 / (double)solver.max_literals);
    if (mem_used != 0) reportf("Memory used           : %.2f MB\n", mem_used / 1048576.0);
    reportf("CPU time              : %g s\n", cpu_time);
}

Solver* solver;
static void SIGINT_handler(int signum) {
    reportf("\n"); reportf("*** INTERRUPTED ***\n");
    printStats(*solver);
    reportf("\n"); reportf("*** INTERRUPTED ***\n");
    exit(1); }


//=================================================================================================
// Main:

void printUsage(char** argv)
{
    reportf("USAGE: %s [options] <input-file> <result-output-file>\n\n  where input may be either in plain or gzipped DIMACS.\n\n", argv[0]);
    reportf("OPTIONS:\n\n");
    reportf("  -polarity-mode = {true,false,rnd}\n");
    reportf("  -decay         = <num> [ 0 - 1 ]\n");
    reportf("  -rnd-freq      = <num> [ 0 - 1 ]\n");
    reportf("  -verbosity     = {0,1,2}\n");
    reportf("\n");
}


const char* hasPrefix(const char* str, const char* prefix)
{
    int len = strlen(prefix);
    if (strncmp(str, prefix, len) == 0)
        return str + len;
    else
        return NULL;
}

/*-------------------------------------------------------------------
 BACKSPACE:


-------------------------------------------------------------------*/
unsigned int getRandom(unsigned int bound){
  
  unsigned int value;
  double r;

  value = bound;
  if(value == 0xFFFFFFFF) {
    printf("ERROR: getRandom is receiving bound = 0 from its caller\n");
    fflush(NULL);
    exit(1);
  }
  
  if(value == 0){
    value = 0;
  } 
  else{
    r = (double)rand() / ((double)(RAND_MAX)+(double)(1));
    value = (unsigned int) (r * value);

  }
  return value;
}
/*-------------------------------------------------------------------
 BACKSPACE:


-------------------------------------------------------------------*/
unsigned int*  readInFile2(const string filename, unsigned int* size){

  ifstream inFile;
  char c;
  vector<int> dest;
  char temp_int[13];
  int i =0;  
  dest.clear();
  inFile.open(filename.c_str());
  if (!inFile) cerr << "Could not open: " << filename.c_str()<<endl;
  
  while(inFile.get(c) && c != '{');
  inFile.get(c);// discard '\n'
  while(inFile.get(c) && c != '}') {
    if(c != ',' && c != '\n'){
      temp_int[i] = c;
      i++;
    }else{
      dest.push_back(atoi(temp_int));
      memset(temp_int,'\n',13);
      i = 0;
    }  
  }

  inFile.close();
  
  //cout << "\n dest.size =" << dest.size() << endl;
  
  unsigned int *ptr = (unsigned int*)malloc(dest.size()*sizeof(unsigned int));
  for (unsigned int i = 0; i < dest.size(); i++) 
    ptr[i] = dest[i];

  *size = dest.size();
  dest.clear();

  return ptr;
}


int main(int argc, char** argv)
{
  Solver      S;
  S.verbosity = 1;
  
  int         i, j;
  const char* value;
  
  float alpha;
  long int seed;
  time_t *tp;
  tp = (time_t*)malloc(sizeof(time_t));
  time(tp);

  
  for (i = j = 0; i < argc; i++){
    if ((value = hasPrefix(argv[i], "-polarity-mode="))){
      if (strcmp(value, "true") == 0)
	S.polarity_mode = Solver::polarity_true;
      else if (strcmp(value, "false") == 0)
	S.polarity_mode = Solver::polarity_false;
      else if (strcmp(value, "rnd") == 0)
	S.polarity_mode = Solver::polarity_rnd;
      else{
	reportf("ERROR! unknown polarity-mode %s\n", value);
	exit(0); }
      
    }else if ((value = hasPrefix(argv[i], "-rnd-freq="))){
      double rnd;
      if (sscanf(value, "%lf", &rnd) <= 0 || rnd < 0 || rnd > 1){
	reportf("ERROR! illegal rnd-freq constant %s\n", value);
	exit(0); }
      S.random_var_freq = rnd;
      
    }else if ((value = hasPrefix(argv[i], "-decay="))){
      double decay;
      if (sscanf(value, "%lf", &decay) <= 0 || decay <= 0 || decay > 1){
	reportf("ERROR! illegal decay constant %s\n", value);
	exit(0); }
      S.var_decay = 1 / decay;
      
    }else if ((value = hasPrefix(argv[i], "-verbosity="))){
      int verbosity = (int)strtol(value, NULL, 10);
      if (verbosity == 0 && errno == EINVAL){
	reportf("ERROR! illegal verbosity level %s\n", value);
	exit(0); }
      S.verbosity = verbosity;
      
    }else if ((value = hasPrefix(argv[i], "-alpha="))){
      sscanf(value, "%f", &alpha);
      if (alpha >= 1){
	reportf("ERROR! alpha unit is percentage of the design. It must be in (0,1): %s\n", value);
	exit(0); 
      }
    }else if ((value = hasPrefix(argv[i], "-seed="))){
      sscanf(value, "%li", &seed);
      if (seed < 0){
	reportf("ERROR! Seed must be gt 0: %s\n", value);
	exit(0); 
      }
      if(seed == 0){
	srand(*tp);
	printf("Random seed = %d\n",(int)*tp);
      }else{
	srand(seed);
	printf("Random seed = %d\n",(int)seed);
      }
      
    }else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "-help") == 0 || strcmp(argv[i], "--help") == 0){
      printUsage(argv);
      exit(0);
      
    }else if (strncmp(argv[i], "-", 1) == 0){
      reportf("ERROR! unknown flag %s\n", argv[i]);
      exit(0);
      
    }else
      argv[j++] = argv[i];
  }
  argc = j;
  
  
  reportf("This is MiniSat 2.0 beta\n");
#if defined(__linux__)
  fpu_control_t oldcw, newcw;
  _FPU_GETCW(oldcw); newcw = (oldcw & ~_FPU_EXTENDED) | _FPU_DOUBLE; _FPU_SETCW(newcw);
  reportf("WARNING: for repeatability, setting FPU to use double precision\n");
#endif
  double cpu_time = cpuTime();
  
  solver = &S;
  signal(SIGINT,SIGINT_handler);
  signal(SIGHUP,SIGINT_handler);
  
  if (argc == 1)
    reportf("Reading from standard input... Use '-h' or '--help' for help.\n");
  
  gzFile in = (argc == 1) ? gzdopen(0, "rb") : gzopen(argv[1], "rb");
  if (in == NULL)
    reportf("ERROR! Could not open file: %s\n", argc == 1 ? "<stdin>" : argv[1]), exit(1);
  
  reportf("============================[ Problem Statistics ]=============================\n");
  reportf("|                                                                             |\n");
  
  parse_DIMACS(in, S);
  gzclose(in);
  FILE* res = (argc >= 3) ? fopen(argv[2], "wb") : NULL;
  
  double parse_time = cpuTime() - cpu_time;
  reportf("|  Parsing time:         %-12.2f s                                       |\n", parse_time);
  
  if (!S.simplify()){
    reportf("Solved by unit propagation\n");
    if (res != NULL) fprintf(res, "UNSAT\n"), fclose(res);
    printf("UNSATISFIABLE\n");
    exit(20);
  }
  /////////////////////////////////////////////////////////////////////////////
  //                      BACKSPACE code starts here                        //
  ////////////////////////////////////////////////////////////////////////////
  unsigned int nlatches, 
    *nlatches_ptr = (unsigned int*) malloc(sizeof(unsigned int));
  unsigned int nsigns, 
    *nsigns_ptr = (unsigned int*) malloc(sizeof(unsigned int));
  unsigned int nconstrs, 
    *nconstrs_ptr = (unsigned int*) malloc(sizeof(unsigned int));
   
  unsigned int* latches = readInFile2("cnfLatches.h", nlatches_ptr);
  nlatches = (*nlatches_ptr);

  
  unsigned int* misat_sign   = readInFile2("constraints.h",nsigns_ptr);
  nsigns = *nsigns_ptr;

  
  unsigned int* constraintSet = readInFile2("constraint_literals.h", nconstrs_ptr);
  nconstrs = *nconstrs_ptr;

  assert((nsigns == nlatches) && "Error: #of(Total latches) != #of(Total Constraints)"); 
  
  
  vec<Lit> BlkClause;
  BlkClause.clear();
  vec<lbool> clause;
  int count = 0;
  unsigned int indx;
  unsigned int indxa, indxb;
  Lit myLit;
  vec<Lit> InitClause;
  InitClause.clear();
  InitClause.growTo(1);
  
  vec<int> MyConstrVars;
  MyConstrVars.clear();
  MyConstrVars.growTo(nlatches);
  for(indx = 0 ; indx<nlatches; indx++) MyConstrVars[indx] = 0;
  unsigned int count_seeds=0;
  unsigned int fraction = (unsigned int)(nlatches*alpha);
  assert(fraction>0 && "Fraction <= 0. Make sure alpha is in (0,1), and nlatches > 0\n");
  
  vec<int> MyRandVect;
  MyRandVect.clear();
  MyRandVect.growTo(nlatches);
  int all  = 0; 
  int none = 0; 
  //   all none
  //    0  0    : some constraints
  //    0  1    : no constraints <=> pre-img set
  //    1  -    : exact one assignment that matches constraint.h (not quite yet, none has to be set to 1 for now)
  
#ifdef BACKSPACE_RANDOM_CONSTRAINTS
  indx=0;
  if(!all){
    for(indx = 0 ; indx<nlatches; indx++) 
      MyRandVect[indx] = 0;
    if(!none){
      while((count_seeds+1) < fraction){
	indx=getRandom(nlatches-1);
	if((MyRandVect[indx] != 1) && (indx != 702)){ /// FMP this line need doc
	  MyRandVect[indx] = 1;
	  count_seeds++;
	  printf("Randomly chosen=%d\n",latches[indx]); 
	}
      }
    }
#endif
    
#ifdef BACKSPACE_VERBOSE
    for(indx = 0 ; indx<nconstrs; indx++)  
      printf("index[%d]-= %d\n", indx,constraintSet[indx]);
#endif
    
    //
    // Requires  constraintSet[] to be in ascending order
    //
    if(!all){
      for(indx = 0 ; indx<nlatches; indx++) 
	MyRandVect[indx] = 0;
      if(!none){
	for(indxa = 0 ; indxa<nconstrs; indxa++) 
	  for(indxb = 0; indxb <nlatches; indxb++)
	    if(latches[indxb] == constraintSet[indxa]){
	      MyRandVect[indxb] = 1;
#ifdef BACKSPACE_VERBOSE
	      printf("Chosen[%d]= %d\n", indxa,constraintSet[indxa]);
#endif
	      count_seeds++;
	    }
      }
    }else{
      for(indx = 0 ; indx<nlatches; indx++) {
	MyRandVect[indx] = 1;
	count_seeds++;
      }
    }
    
    printf("Total # of latches chosen =%d\n",count_seeds);
    
    //
    // Add initial clauses (constraints) to the SAT instance
    //
    for(indx=0; indx<nlatches;indx++){
      if(MyRandVect[indx]){
	MyConstrVars[indx] = 1;
	InitClause[0] = misat_sign[indx] == 1 ? Lit(latches[indx]-1) : ~Lit(latches[indx]-1);
	S.addClause(InitClause); 
      }
    }
    
    //
    // Main Loop
    //
    FILE* BPhandler;
    BPhandler = fopen("../breakpoint","w");
    bool ret;
    while(ret=S.solve()){
      // Write out breakpoint
      // This for-loop is dependent on a previou techmap where Q and QN 
      // were generated; If using the released techmap, this should be removed
      for (unsigned int i = 0; i< nlatches; i++)
	if ((S.model[latches[i]-1] != l_Undef) && 
	    (latches[i] != 232) && 
	    (latches[i] != 234) && 
	    (latches[i] != 236) && 
	    (latches[i] != 238) && 
	    (latches[i] != 240) && 
	    (latches[i] != 242) && 
	    (latches[i] != 244) && 
	    (latches[i] != 246) && 
	    (latches[i] != 248) && 
	    (latches[i] != 250) && 
	    (latches[i] != 252) && 
	    (latches[i] != 254) && 
	    (latches[i] != 256) && 
	    (latches[i] != 258) && 
	    (latches[i] != 260) && 
	    (latches[i] != 262) && 
	    (latches[i] != 264) && 
	    (latches[i] != 266) && 
	    (latches[i] != 268) && 
	    (latches[i] != 270) && 
	    (latches[i] != 272) && 
	    (latches[i] != 274) && 
	    (latches[i] != 276) && 
	    (latches[i] != 278) && 
	    (latches[i] != 280) && 
	    (latches[i] != 282) && 
	    (latches[i] != 284) && 
	    (latches[i] != 286) && 
	    (latches[i] != 298) && 
	    (latches[i] != 300) && 
	    (latches[i] != 302) && 
	    (latches[i] != 304) && 
	    (latches[i] != 306) && 
	    (latches[i] != 308) && 
	    (latches[i] != 310) && 
	    (latches[i] != 312) && 
	    (latches[i] != 314) && 
	    (latches[i] != 316) && 
	    (latches[i] != 318) && 
	    (latches[i] != 320) && 
	    (latches[i] != 322) && 
	    (latches[i] != 324) && 
	    (latches[i] != 326) && 
	    (latches[i] != 328) && 
	    (latches[i] != 330) && 
	    (latches[i] != 332) && 
	    (latches[i] != 334) && 
	    (latches[i] != 336) && 
	    (latches[i] != 338) && 
	    (latches[i] != 340) && 
	    (latches[i] != 342) && 
	    (latches[i] != 344) && 
	    (latches[i] != 346) && 
	    (latches[i] != 348) && 
	    (latches[i] != 350) && 
	    (latches[i] != 352) && 
	    (latches[i] != 354) && 
	    (latches[i] != 3682) && 
	    (latches[i] != 3712) && 
	    (latches[i] != 3729) && 
	    (latches[i] != 3738) && 
	    (latches[i] != 15535) && 
	    (latches[i] != 31430) && 
	    (latches[i] != 31383) && 
	    (latches[i] != 31381) && 
	    (latches[i] != 35522) && 
	    (latches[i] != 35354) && 
	    (latches[i] != 35387) && 
	    (latches[i] != 35375) && 
	    (latches[i] != 35381) && 
	    (latches[i] != 35594) && 
	    (latches[i] != 35364) && 
	    (latches[i] != 36080) && 
	    (latches[i] != 36173) && 
	    (latches[i] != 36178) && 
	    (latches[i] != 36183) && 
	    (latches[i] != 36188) && 
	    (latches[i] != 36193) && 
	    (latches[i] != 37306)
	    )
	  fprintf(BPhandler,"%s", (S.model[latches[i]-1]==l_True)?"1":"0");
	else if(
		(latches[i] == 232) || 
		(latches[i] == 234) || 
		(latches[i] == 236) || 
		(latches[i] == 238) || 
		(latches[i] == 240) || 
		(latches[i] == 242) || 
		(latches[i] == 244) || 
		(latches[i] == 246) || 
		(latches[i] == 248) || 
		(latches[i] == 250) || 
		(latches[i] == 252) || 
		(latches[i] == 254) || 
		(latches[i] == 256) || 
		(latches[i] == 258) || 
		(latches[i] == 260) || 
		(latches[i] == 262) || 
		(latches[i] == 264) || 
		(latches[i] == 266) || 
		(latches[i] == 268) || 
		(latches[i] == 270) || 
		(latches[i] == 272) || 
		(latches[i] == 274) || 
		(latches[i] == 276) || 
		(latches[i] == 278) || 
		(latches[i] == 280) || 
		(latches[i] == 282) || 
		(latches[i] == 284) || 
		(latches[i] == 286) || 
		(latches[i] == 298) || 
		(latches[i] == 300) || 
		(latches[i] == 302) || 
		(latches[i] == 304) || 
		(latches[i] == 306) || 
		(latches[i] == 308) || 
		(latches[i] == 310) || 
		(latches[i] == 312) || 
		(latches[i] == 314) || 
		(latches[i] == 316) || 
		(latches[i] == 318) || 
		(latches[i] == 320) || 
		(latches[i] == 322) || 
		(latches[i] == 324) || 
		(latches[i] == 326) || 
		(latches[i] == 328) || 
		(latches[i] == 330) || 
		(latches[i] == 332) || 
		(latches[i] == 334) || 
		(latches[i] == 336) || 
		(latches[i] == 338) || 
		(latches[i] == 340) || 
		(latches[i] == 342) || 
		(latches[i] == 344) || 
		(latches[i] == 346) || 
		(latches[i] == 348) || 
		(latches[i] == 350) || 
		(latches[i] == 352) || 
		(latches[i] == 354) || 
		(latches[i] == 3682) || 
		(latches[i] == 3712) || 
		(latches[i] == 3729) || 
		(latches[i] == 3738) || 
		(latches[i] == 15535) || 
		(latches[i] == 31430) || 
		(latches[i] == 31383) || 
		(latches[i] == 31381) || 
		(latches[i] == 35522) || 
		(latches[i] == 35354) || 
		(latches[i] == 35387) || 
		(latches[i] == 35375) || 
		(latches[i] == 35381) || 
		(latches[i] == 35594) || 
		(latches[i] == 35364) || 
		(latches[i] == 36080) || 
		(latches[i] == 36173) || 
		(latches[i] == 36178) || 
		(latches[i] == 36183) || 
		(latches[i] == 36188) || 
		(latches[i] == 36193) || 
		(latches[i] == 37306)
		)
	  fprintf(BPhandler,"%s", (S.model[latches[i]-1]==l_True)?"0":"1");
      fprintf(BPhandler,"\n");
      
      //
      // Print Model
      //
#ifdef BACKSPACE_VERBOSE
      for (unsigned int i = 0; i< nlatches; i++)
	if (S.model[latches[i]-1] != l_Undef)
	  reportf("%s%s%d", (i==0)?"":" ", (S.model[latches[i]-1]==l_True)?" ":"-", latches[i]);
      reportf("\n");
#endif
      
      BlkClause.clear();
      count++;
      if( count % BACKSPACE_PRINT_STEP == 0) {reportf("Iteration %d :\n", count); /*exit(20);*/}
      if( count > BACKSPACEABILITY) {    
	printStats(S);
	reportf("Found %d : Reached upper-bound, exiting...\n", count); break; 
      }
      S.model.copyTo(clause);
      if(nlatches > (unsigned int) clause.size()) {
	reportf ("ERROR Numb. of latches gt clause size : %d > %d\n", nlatches, clause.size());
	exit(20);
      }
      
      for (indx = 0 ; indx< nlatches; indx++){
	myLit = clause[latches[indx]-1] == l_True ? ~Lit(latches[indx]-1) : Lit(latches[indx]-1) ;
	BlkClause.push(myLit);
      }
      if(BlkClause.size() > 0)
	S.addClause(BlkClause);
      else{
	reportf("ERROR: No Blocking Clause generated. Exiting...\n");
	exit(20);
      }
      
    }//while(ret=S.solve())
    fclose(BPhandler);
   /////////////////////////////////////////////////////////////////////////////
   //                      BACKSPACE code ends  here                          //
   ////////////////////////////////////////////////////////////////////////////

    
    printStats(S);
    reportf("\n");
    printf(ret ? "SATISFIABLE\n" : "UNSATISFIABLE\n");
    // BACKSPACE code added
    if( count >= 2) {    
      printStats(S);
      reportf("Found %d : Multiple assignments...\n", count);
    }
    else
      reportf("Found %d satisfying assignments\n",count);
    // BACKSPACE code done
    
    if (res != NULL){
      if (ret){
	fprintf(res, "SAT\n");
	for (int i = 0; i < S.nVars(); i++)
	  if (S.model[i] != l_Undef)
	    fprintf(res, "%s%s%d", (i==0)?"":" ", (S.model[i]==l_True)?"":"-", i+1);
	fprintf(res, " 0\n");
      }else
	fprintf(res, "UNSAT\n");
      fclose(res);
    }
    
#ifdef NDEBUG
    exit(ret ? 10 : 20);     // (faster than "return", which will invoke the destructor for 'Solver')
#endif
  }
  
